%% Author: Wouter Verbeke
% Data Analytics Laboratory
% Department of Business Technology and Operations 
% Faculty of Economic and Social Sciences and Solvay Business School
% Vrije Universiteit Brussel, Pleinlaan 2, B-1050 Brussel, Belgium

%% Documentation
% This file contains the base code for executing the RULEM postprocessing
% approach for making rulesets for ordinal classification to comply with
% monotonicity constraints, as presented in the article:
% Wouter Verbeke, David Martens, Bart Baesens
% RULEM: A novel heuristic rule learning approach for ordinal classification with monotonicity constraints
% Applied Soft Computing, In press, 2017

function [baseAcc,basePos,compAcc,compPos,rulesetTotal,RUMS,RUMF,NP,WC,baseMAE,baseMSE,compMAE,compMSE]  = ...
             RULEM(ruleset,constraints,varType,trainData,testData,adjValue,...
             NPWCparam,adjBorder,defaultRule)

%% ------------------------------------------------------------------------
%% WARNING: class label should be in the last column of the data!
%% ------------------------------------------------------------------------
%% EXTENSION: optimize the additional rules merging algorithm
%% EXTENSION: optimize procedure to reach NPWC = 0 in fewer steps ( = fewer 
%%            additional rules)
%% ------------------------------------------------------------------------
%% EXTENSION: take into account wrong classified instances in cells?
%% EXTENSION: allow more flexibility by setting parameter to tune prop.
%%            overwriting existing rules/creating new rules to resolve
%%            violations
%% EXTENSION: allow more flexibility by setting minimum sens/spec/acc that
%%            final ruleset has to reach: is sort of soft constraint - or 
%%            maximum number of additional rules - or minimum justifiab.
%%            in terms of RULEMS or RULEMF
%% ------------------------------------------------------------------------

%% ------------------------------------------------------------------------
%% General information:
%% - ruleset format: 
%%      columns: variables:
%%          ordinal variables: two values (column) 
%%          nominal variables: one value (column)
%%          last column:       class label
%%      rows: rules:
%%          ordinal variables: lower border: equal to or larger than x 
%%                                           (larger than value x: is 
%%                                           replaced by >= x + adjValue 
%%                                           (with adjValue ~ eps))
%%                             upper border: smaller than x (equal to or 
%%                                           smaller than value x: is 
%%                                           replaced by < x + adjValue 
%%                                           (with adjValue ~ eps))
%%          nominal variables: equal to x 
%%                             (with x = anyValue indicating/comprising all
%%                             possible values, and anyValue implicitly 
%%                             passed as argument within defaultRule)
%% - varType: 0 = nominal variable, 1 = ordinal variable
%% - trainData/testData: the data used to train and evaluate the ruleset
%% - constraints: -1: negative constr., 0: no constraint, 1 positive 
%%   constraint
%% - adjValue has to be smaller than smallest difference between two values
%%   of any attribute in the data set
%% - NPWC param: default value: 0.5
%% - adjBorder: indicate whether ruleset is in correct format, or whether
%%   upper boundaries have to be adjusted (from =< to < by adding adjValue)
%% - defaultRule: if the last rule in the ruleset is not a defaultrule, 
%%   e.g. when ruleset results from decision tree (J48), then a defaultrule
%%   should be explicitly provided as an argument
%% ------------------------------------------------------------------------

%% Manual input: simple example:
% ruleset = [ 1 3 3 4 2;
%             0 2 1 2 1;
%             0 3 0 4 0];
% constraints = [1 1];
% varType = [1 1];
% trainData = [1.5 1.5 1
%              1.5 3.5 2
%              1.5 2.2 1;
%              1.5 2.8 2];
% testData = [0.5 3 1;
%             1.5 2.1 1;
%             1.5 2.7 2;
%             2.5 0.5 0;
%             1 1.5 1;
%             2 3.5 2];
% adjValue = 0.01;
% NPWCparam = 0.5;
% adjBorder = 1;
% defaultRule = ruleset(end,:);
% meanNPmax = 0.5;
NPWCmax = 50;
% disp(['            Maximum fraction corrupted cells parameter: ' num2str(meanNPmax)]);
disp(['            Maximum NPWC parameter: ' num2str(NPWCmax)]);

%% Derive parameters from input
if nargin < 7
    NPWCparam = 0.5;
end
if nargin < 8
    adjBorder = 0;
end
if nargin < 9
    defaultRule = ruleset(end,:);
end
dataClassIdx = size(trainData,2); % the class attribute should be the last in the data
rulesetClassIdx = size(ruleset,2); % the class label is assigned by the last column in the rule set
classLabels = sort(unique(trainData(:,dataClassIdx)));
nrClasses = size(classLabels,1);
nrVar = size(constraints,2);
varIndex = zeros(1,nrVar);
index = 1;
anyValue = []; % init: if no nominal attributes
for i = 1:nrVar
    varIndex(1,i) = index;
    if varType(1,i) == 0
        anyValue = defaultRule(1,index); % the anyValue indicator is the same for each nominal variable, and is the value of the nominal variables in the defaultRule
        index = index + 1;
    else
        index = index + 2;
    end
end

%% Adjust borders of rulesetCurrent to format of RUMII if required
if adjBorder
    varOrdRight = varIndex(varType==1)+1;
    ruleset(:,varOrdRight) = ruleset(:,varOrdRight) + adjValue;
end

%% Initialize the n-dimensional grid, with n the number of attributes
disp('            Initializing grid');
nrCuts = zeros(1,nrVar);
cutVals = cell(1,nrVar);
nrCellsVar = zeros(1,nrVar);
for i = 1:nrVar
    if varType(1,i) == 0
        cutVals{1,i} = unique(ruleset(:,varIndex(1,i)));
        cutVals{1,i} = setdiff(cutVals{1,i},anyValue); % anyValue is not a separate cut of the grid for a nominal attribute
        if isempty(cutVals{1,i}) % then nominal attribute always takes value = anyValue
            cutVals{1,i}= anyValue;
        end
        nrCellsVar(1,i) = size(cutVals{1,i},1);
    elseif varType(1,i) == 1
        cutVals{1,i} = unique([ruleset(:,varIndex(1,i));ruleset(:,varIndex(1,i)+1)]);
        nrCellsVar(1,i) = size(cutVals{1,i},1)-1;
    end
    nrCuts(1,i) = size(cutVals{1,i},1);
end
cuts = zeros(max(nrCuts),nrVar);
for i = 1:nrVar
    cuts(1:nrCuts(1,i),i) =  cutVals{1,i}(:,1);
end
grid = zeros(nrCellsVar);
% Label each cell in the grid
nrCells = prod(nrCellsVar);
indices = ind2sub2(nrCellsVar,(1:+1:nrCells));
middleCell = zeros(1,nrVar);
for i = 1:nrCells
    for j = 1:nrVar % compose center of the cell to find label
        if varType(1,j) == 0
            middleCell(1,j) = cuts(indices(i,j),j);
        else
            middleCell(1,j) = (cuts(indices(i,j),j) + cuts(indices(i,j)+1,j))/2;
        end
    end
    grid(sub2ind2(size(grid),num2cell(indices(i,:)))) = applyRuleset(ruleset,middleCell,varType,anyValue,rulesetClassIdx);
end

%% Calculate NP, WC and compose NPWC
disp('            Calculating NP/WC grid');
[NP,gridNP,~,~,~,~,meanNP] = calculateGridNP(grid,nrCellsVar,constraints);
[WC gridWC] = calculateGridWC(grid,nrCellsVar,constraints,varType,anyValue,trainData,cuts);
NPWC = (NPWCparam*NP + (1-NPWCparam)*WC);
gridNPWC = (NPWCparam*gridNP + (1-NPWCparam)*gridWC);
disp(['                Start NPWC = ' num2str(NPWC)]);
% disp(['                Mean NP = ' num2str(meanNP)]);
[RUMS RUMF] = calculateJustifiability(gridNP,cuts,nrCellsVar,defaultRule,[trainData;testData],varType);
disp(['                RUMS = ' num2str(RUMS)]);
disp(['                RUMF = ' num2str(RUMF)]);

%% Resolve violations by adjusting labels: minimization of the NPWC:
noImpCounter = 0;
maxNoImpCounter = 10;
if NPWC > 0 && NPWC <= NPWCmax %&& meanNP <= meanNPmax
    disp('            Resolving violations');
    rulesAdd = [];
    indexList = [];
    while NPWC > 0 && noImpCounter < maxNoImpCounter
        % find cell with maximum NPWC value - if multiple cells are found with same max value - take first one
        noImp = 1;
        sortedNPWC = totalSort(gridNPWC,'descend');
        mult = 0;
        counter = 1;
        while noImp && noImpCounter < maxNoImpCounter
            if mult
                multInd = multInd + 1;
                index = indexTotal(multInd,1);
                counter = counter + 1;
                if multInd == numel(indexTotal)
                    mult = 0;
                end
            else
                indexTotal = find(gridNPWC == sortedNPWC(counter));
                if numel(indexTotal) > 1
                    mult = 1;
                    multInd = 1;
                end
                index = indexTotal(1);
                counter = counter + 1;
            end
            % relabel cell to minimize NP
            orLabel = grid(index);
            NPWCalt = zeros(1,nrClasses);
            for i = 1:nrClasses
                if classLabels(i,1) == orLabel
                    NPWCalt(1,i) = NPWC;
                else    
                    grid(index) = classLabels(i,1);
                    NPalt = calculateGridNP(grid,nrCellsVar,constraints);
                    WCalt = calculateGridWC(grid,nrCellsVar,constraints,varType,anyValue,trainData,cuts);
                    NPWCalt(1,i) = (NPWCparam*NPalt + (1-NPWCparam)*WCalt);
                end
            end
            [NPWCnew,label] = min(NPWCalt);
            if classLabels(label,1) ~= orLabel && NPWCnew < NPWC
                rulesAdd = [ind2sub2(nrCellsVar,index) classLabels(label,1);rulesAdd];
                grid(index) = classLabels(label,1);
                [NP gridNP] = calculateGridNP(grid,nrCellsVar,constraints);
                [WC gridWC] = calculateGridWC(grid,nrCellsVar,constraints,varType,anyValue,trainData,cuts);
                NPWC = (NPWCparam*NP + (1-NPWCparam)*WC);
                gridNPWC = (NPWCparam*gridNP + (1-NPWCparam)*gridWC);
                noImp = 0;
                disp(['                NPWC = ' num2str(NPWC)]);
                indexList = [indexList;index];
            else
                grid(index) = orLabel;
            end
            if counter > numel(sortedNPWC) % then no additional rule improves NPWC: cell with highest NPWC is adjusted
                disp('                no improvement of NPWC - cell with highest NPWC selected');
                noImpCounter = noImpCounter + 1;
                indexTotal = find(gridNPWC == sortedNPWC(1));
                index = indexTotal(1);
                orLabel = grid(index);
                NPWCalt = zeros(1,nrClasses);
                for i = 1:nrClasses
                    if classLabels(i,1) == orLabel
                        NPWCalt(1,i) = NPWC;
                    else    
                        grid(index) = classLabels(i,1);
                        NPalt = calculateGridNP(grid,nrCellsVar,constraints);
                        WCalt = calculateGridWC(grid,nrCellsVar,constraints,varType,anyValue,trainData,cuts);
                        NPWCalt(1,i) = (NPWCparam*NPalt + (1-NPWCparam)*WCalt);
                    end
                end
                [B,IX] = sort(NPWCalt,'descend');
%                 [~,label] = min(NPWCalt);
                label = IX(2);
                rulesAdd = [ind2sub2(nrCellsVar,index) classLabels(label,1); rulesAdd];
                grid(index) = classLabels(label,1);
                [NP gridNP] = calculateGridNP(grid,nrCellsVar,constraints);
                [WC gridWC] = calculateGridWC(grid,nrCellsVar,constraints,varType,anyValue,trainData,cuts);
                NPWC = (NPWCparam*NP + (1-NPWCparam)*WC);
                gridNPWC = (NPWCparam*gridNP + (1-NPWCparam)*gridWC);
                noImp = 0;
                disp(['                NPWC = ' num2str(NPWC)]);
                indexList = [indexList;index];
            end
        end
    end
   
    %% Replace cell indices in rulesAdd by cell boundaries 
    nrRulesAdd = size(rulesAdd,1);
    rulesetAdd = zeros(nrRulesAdd,size(ruleset,2));
    for j = 1:nrVar
        rulesetAdd(:,varIndex(1,j)) = cuts(rulesAdd(:,j),j);
        if varType(1,j) == 1
            rulesetAdd(:,varIndex(1,j)+1) = cuts(rulesAdd(:,j)+1,j);
        end
    end
    rulesetAdd(:,rulesetClassIdx) = rulesAdd(:,end);
    [~,indUni] = unique(rulesAdd(:,1:end-1),'rows','first');
    indUni = sort(indUni,'ascend');
    rulesAdd = rulesAdd(indUni,:);
    rulesetAdd = rulesetAdd(indUni,:);

    %% Minimize nr of additional rules by merging neighboring cell-rules
    disp('            Merging added rules - To be optimized further?');
    disp(['                Initial nr. of additional rules: ' num2str(size(rulesetAdd,1))]);
    rulesetAdd = mergeCells(rulesAdd,rulesetAdd,cuts,varType);
    disp(['                Final nr. of additional rules: ' num2str(size(rulesetAdd,1))]);
    rulesetTotal = [rulesetAdd; ruleset];
    [~,indUni] = unique(rulesetTotal(:,1:end-1),'rows','first');
    indUni = sort(indUni,'ascend');
    rulesetTotal = rulesetTotal(indUni,:);
    disp(['                Initial nr. of rules: ' num2str(size(ruleset,1))]);
    disp(['                Final nr. of rules: ' num2str(size(rulesetTotal,1))]);
    
    %% Calculate performance indices of resulting ruleset:
    disp('            Calculating performance indices');
    [baseAcc,basePos,~,~,baseMAE,baseMSE] = evalRulesetMaeMse(ruleset,testData,varType,anyValue,dataClassIdx,rulesetClassIdx);
    [compAcc,compPos,~,~,compMAE,compMSE] = evalRulesetMaeMse(rulesetTotal,testData,varType,anyValue,dataClassIdx,rulesetClassIdx);
    disp(['                Base accuracy = ' num2str(baseAcc)]);
    disp(['                Base positive = ' num2str(basePos)]);
    disp(['                Comp accuracy = ' num2str(compAcc)]);
    disp(['                Comp positive = ' num2str(compPos)]);
    disp(['                Base MAE = ' num2str(baseMAE)]);
    disp(['                Base MSE = ' num2str(baseMSE)]);
    disp(['                Comp MAE = ' num2str(compMAE)]);
    disp(['                Comp MSE = ' num2str(compMSE)]);

    %% Check whether final NPWC is zero:
    [NP,WC]  = calcNPWC(rulesetTotal,constraints,varType,trainData);
    NPWC = (NPWCparam*NP + (1-NPWCparam)*WC);
    disp(['            Final monotonicity check: NPWC = ' num2str(NPWC)]);

%     %% Calculation of RUMS and RUMF
%     data = [trainData; testData];
%     RUMS = 1;
%     nrValues = zeros(1,size(varType,2));
%     for i = 1:size(varType,2)
%         if varType(1,i) == 0
%             nrValues(1,i) = numel(unique(data(:,i)));
%         end
%     end
%     for i = 1:size(rulesetAdd,1)
%         RUMS = RUMS - ruleSpace(rulesetAdd(i,:),defaultRule,varType,nrValues);
%     end
%     RUMF = 1 - (size(coveredInstances(rulesetAdd,data,varType,anyValue),1))/size(data,1);
    
%% If no violations
elseif NPWC > 0 &&  NPWC > NPWCmax % || meanNP > meanNPmax
%     if meanNP > meanNPmax && NPWC > NPWCmax
%         disp('            Maximum fraction of corrupted cells and maximum NPWC reached');
%         rulesetTotal = -1;
%         compAcc = -1;
%         compPos = -1;
%         RUMS = -1;
%         RUMF = -1;
%     elseif NPWC > NPWCmax
        disp('            Maximum NPWC reached');
        rulesetTotal = -2;
        compAcc = -2;
        compPos = -2;
        RUMS = -2;
        RUMF = -2;
%     else
%         disp('            Maximum fraction of corrupted cells reached');
%         rulesetTotal = -3;
%         compAcc = -3;
%         compPos = -3;
%         RUMS = -3;
%         RUMF = -3;
%     end
    [baseAcc, basePos,~,~,baseMAE,baseMSE] = evalRulesetMaeMse(ruleset,testData,varType,anyValue,dataClassIdx,rulesetClassIdx);
    compAcc = -1;
    compPos = -1;
    RUMS = -1;
    RUMF = -1;
    compMAE = -1;
    compMSE = -1;
else
    disp('            No violations, calculating performance');
    rulesetTotal = ruleset;
    [baseAcc, basePos,~,~,baseMAE,baseMSE] = evalRulesetMaeMse(ruleset,testData,varType,anyValue,dataClassIdx,rulesetClassIdx);
    compAcc = baseAcc;
    compPos = basePos;
    compMAE = baseMAE;
    compMSE = baseMSE;
    disp(['                Base accuracy = ' num2str(baseAcc)]);
    disp(['                Base positive = ' num2str(basePos)]);
    disp(['                Comp accuracy = ' num2str(compAcc)]);
    disp(['                Comp positive = ' num2str(compPos)]);
    disp(['                Base MAE = ' num2str(baseMAE)]);
    disp(['                Base MSE = ' num2str(baseMSE)]);
    disp(['                Comp MAE = ' num2str(compMAE)]);
    disp(['                Comp MSE = ' num2str(compMSE)]);
end