%% Author: Wouter Verbeke
% Data Analytics Laboratory
% Department of Business Technology and Operations 
% Faculty of Economic and Social Sciences and Solvay Business School
% Vrije Universiteit Brussel, Pleinlaan 2, B-1050 Brussel, Belgium

%% Documentation
% This file contains the code for running the benchmarking experiments 
% published in the above mentioned paper.
% The experiments postprocess rulesets induced by Weka and AntMiner+ 

%% Citation:
% Wouter Verbeke, David Martens, Bart Baesens
% RULEM: A novel heuristic rule learning approach for ordinal classification with monotonicity constraints
% Applied Soft Computing, In press, 2017

%% Experimental data and parameters controlling the experiments
NPWCparam = 0.5;
makeInputfiles = 1; % for cross-validation
nrSplits = 10; % number of hold out sets
beginSplit = 1;
endSplit = 10;
beginDataset = 1;
endDataset = 14;
RipperPos = 1;
J48Pos = 2;
AntMinerPos = 3;
classifiers = {'rules.JRip';'trees.J48';'AntMiner+'};
adjValue = 0.0001; % should be checked whether small enough
trainFraction = 0.66; % fraction of data used to train model
homefolder = 'C:\Matlab\RULEM\';
dataFiles = {
'auto.v.1','auto.train','auto.test','auto.v.2';
'balance.v.1','balance.train','balance.test','balance.v.2';
'breast-cancer.v.1','breast-cancer.train','breast-cancer.test','breast-cancer.v.2';
'car.v.1','car.train','car.test','car.v.2';
'churn.v.1','churn.train','churn.test','churn.v.2';
'cmc.v.1','cmc.train','cmc.test','cmc.v.2';
'era.v.1','era.train','era.test','era.v.2';
'esl.v.1','esl.train','esl.test','esl.v.2';
'german.v.1','german.train','german.test','german.v.2';
'haberman.v.1','haberman.train','haberman.test','haberman.v.2';
'housing.v.1','housing.train','housing.test','housing.v.2';
'lev.v.1','lev.train','lev.test','lev.v.2';
'pima.v.1','pima.train','pima.test','pima.v.2';
'swd.v.1','swd.train','swd.test','swd.v.2';
};

%% Create random holdout train/test files for cross-validation
if makeInputfiles
    for i = beginDataset:endDataset 
        wekaObj = loadARFF(strcat(homefolder,dataFiles{i,1},'.arff'));
        [data,featureNames,targetNDX,stringVals,relationName] = weka2matlab(wekaObj);
        for j = beginSplit:endSplit
            data = shuffle(data);
            nrInstances = size(data,1);
            traindata = data(1:round(trainFraction*nrInstances),:);
            testdata = data(round(trainFraction*nrInstances)+1:end,:);
            matlab2arff(traindata,strcat(homefolder,dataFiles{i,1},'.arff'),strcat(homefolder,dataFiles{i,2},'.',num2str(j),'.arff'));
            matlab2arff(testdata,strcat(homefolder,dataFiles{i,1},'.arff'),strcat(homefolder,dataFiles{i,3},'.',num2str(j),'.arff'));
            % AntMiner+ shuffles:
            matlab2arff(data,strcat(homefolder,dataFiles{i,4},'.arff'),strcat(homefolder,dataFiles{i,4},'.',num2str(j),'.','1','.txt'));
        end
    end
end

%% Constraints
%  0: no constraint
%  1: positive constraint
% -1: negative constraint
constraints = {
[0 1 0 1 0 -1 0];                                       %auto
[-1 -1 1 1];                                            %balance
[0 0 1 1 0 1 0 0 0];                                    %breast-cancer
[1 1 1 1 1 1];                                          %car
[0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 -1];               %churn
[0 0 0 1 0 0 1 0];                                      %contraceptive  
[1 1 1 1];                                              %era
[1 1 1 1];                                              %esl    
[-1 1 -1 0 0 -1 0 0 0 0];                               %german2
[1 -1 1];                                               %haberman
[0 0 0 0 0 0 0 0 0 0 0 0 -1];                           %housing
[1 1 1 1];                                              %lev  
[1 0 0 0 0 1 1 1];                                      %pima
[1 1 1 1 1 1 1 1 1 1];                                  %swd
};

%% Variable types (see also readData.m)
%  0: nominal
%  1: ordinal
variableType = {
[1 1 1 1 1 1 1];                                        %auto
[1 1 1 1];                                              %balance
[1 1 1 1 1 1 1 1 1];                                    %breast-cancer
[1 1 1 1 1 1];                                          %car
[1 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1];                %churn
[1 1 1 1 1 1 1 1];                                      %contraceptive
[1 1 1 1];                                              %era
[1 1 1 1];                                              %esl
[1 1 1 1 1 1 1 1 1 1];                                  %german2
[1 1 1];                                                %haberman
[1 1 1 1 1 1 1 1 1 1 1 1 1];                            %housing
[1 1 1 1];                                              %lev
[1 1 1 1 1 1 1 1];                                      %pima
[1 1 1 1 1 1 1 1 1 1];                                  %swd
};

%% Predefining result matrices - allocating:
ruleset = cell(size(dataFiles,1),size(classifiers,1),nrSplits);
rulesetTotal = ruleset;
baseAcc = zeros(size(dataFiles,1),size(classifiers,1),nrSplits);
basePos = baseAcc;
compAcc = baseAcc;
compPos = baseAcc;
baseMAE = baseAcc;
baseMSE = baseAcc;
compMAE = baseAcc;
compMSE = baseAcc;
fraction = baseAcc;
nrCoveredInstances = baseAcc;

%% Experiments:
tic
for k = beginSplit:endSplit % for all holdout sets
    disp(['Hold out split: ' num2str(k)]);
    for i = beginDataset:endDataset 
        disp(['  Dataset: ' num2str(i)]);
        wekaTrainingObj = loadARFF(strcat(homefolder,dataFiles{i,2},'.',num2str(k),'.arff'));
        wekaTestObj = loadARFF(strcat(homefolder,dataFiles{i,3},'.',num2str(k),'.arff'));
        [traindata,featureNames,targetNDX,stringVals,relationName] = weka2matlab(wekaTrainingObj);
        testdata = weka2matlab(wekaTestObj);
        % Ripper
        if RipperPos > 0
            disp('       Ripper')
            wekaClassifier = trainWekaClassifier(wekaTrainingObj,classifiers{RipperPos,1});
            modelString = char(wekaClassifier.toString);
            ruleset{i,RipperPos,k} = getRulesetRipper(modelString,variableType{i,1},[traindata;testdata],featureNames);
            [baseAcc(i,RipperPos,k),basePos(i,RipperPos,k),compAcc(i,RipperPos,k),compPos(i,RipperPos,k),rulesetTotal{i,RipperPos,k},fraction(i,RipperPos,k),nrCoveredInstances(i,RipperPos,k),baseMAE(i,RipperPos,k),baseMSE(i,RipperPos,k),compMAE(i,RipperPos,k),compMSE(i,RipperPos,k)] = RULEM(ruleset{i,RipperPos,k},constraints{i,1},variableType{i,1},traindata,testdata,adjValue,NPWCparam,1);
        end
        % J48
        if J48Pos > 0
            disp('       J48')
            wekaClassifier = trainWekaClassifier(wekaTrainingObj,classifiers{J48Pos,1});
            modelString = char(wekaClassifier.toString);
            [ruleset{i,J48Pos,k} defaultRule] = getRulesetJ48(modelString,variableType{i,1},[traindata;testdata],featureNames,adjValue);
            [baseAcc(i,J48Pos,k),basePos(i,J48Pos,k),compAcc(i,J48Pos,k),compPos(i,J48Pos,k),rulesetTotal{i,J48Pos,k},fraction(i,J48Pos,k),nrCoveredInstances(i,J48Pos,k),baseMAE(i,RipperPos,k),baseMSE(i,RipperPos,k),compMAE(i,RipperPos,k),compMSE(i,RipperPos,k)] = RULEM(ruleset{i,J48Pos,k},constraints{i,1},variableType{i,1},traindata,testdata,adjValue,NPWCparam,0,defaultRule);
        end
        % AntMiner+
        if AntMinerPos > 0
            disp('       AntMiner+')
            input = strcat(homefolder,dataFiles{i,4},'.',num2str(k),'.');
            output = strcat(homefolder,'AntMiner+\');
            procedure = 'H -T 1';
            options = '';
            [finalAccuracy, decisionList, majorityClass, anyValue, nbValues, traindataAntMiner, testdataAntMiner, list] = AntminerADJ(input, output,procedure,options);
            disp(['accuracy AntMiner ruleset: ' num2str(finalAccuracy)])
            defaultRule = decisionList(1,:); % create defaultRule and add to decisionList = ruleset
            varIndex = 1;
            for j = 1:size(variableType{i,1},2)
                if variableType{i,1}(1,j) == 0
                    defaultRule(1,varIndex) = anyValue;
                    varIndex = varIndex+1;
                else
                    defaultRule(1,varIndex) = 1;
                    defaultRule(1,varIndex+1) = nbValues(1,j+2); %+2 since alpha and beta parameters are included in nbValues
                    varIndex = varIndex+2;
                end
            end
            defaultRule(1,end) = majorityClass;
            ruleset{i,AntMinerPos,k} = [decisionList;defaultRule];
            [baseAcc(i,AntMinerPos,k),basePos(i,AntMinerPos,k),compAcc(i,AntMinerPos,k),compPos(i,AntMinerPos,k),rulesetTotal{i,AntMinerPos,k},fraction(i,AntMinerPos,k),nrCoveredInstances(i,AntMinerPos,k),baseMAE(i,RipperPos,k),baseMSE(i,RipperPos,k),compMAE(i,RipperPos,k),compMSE(i,RipperPos,k)] = RULEM(ruleset{i,AntMinerPos,k},constraints{i,1},variableType{i,1},double(traindataAntMiner),double(testdataAntMiner),adjValue,NPWCparam,1);
        end
    end
end
experimentTime = toc;

%% Calculate the number of rules and the average number of tests per rule 
%% in each ruleset
nrTestsRule = zeros(size(dataFiles,1),size(classifiers,1),nrSplits);
nrRules = nrTestsRule;
nrTestsRuleTotal = nrTestsRule;
nrRulesTotal = nrTestsRule;
for i = 1:14
    for j = 1:3
        for k = 1:5
            if ~isempty(ruleset{i,j,k})
                if j == AntMinerPos || j == RipperPos % ignore defaultrule for attribute tests
                    nrTestsRuleArray = countTestsRuleset(ruleset{i,j,k}(1:end-1,:),ruleset{i,j,k}(end,:),variableType{i,1});
                    nrTestsRule(i,j,k) = mean(nrTestsRuleArray,1);
                    nrRules(i,j,k) = size(ruleset{i,j,k},1);
                else % if j == J48Pos: use defaultRule of ripper, with adjValue
                    nrTestsRuleArray = countTestsRuleset(ruleset{i,j,k}(1:end,:),ruleset{i,RipperPos,k}(end,:),variableType{i,1},adjValue);
                    nrTestsRule(i,j,k) = mean(nrTestsRuleArray,1);
                    nrRules(i,j,k) = size(ruleset{i,j,k},1);
                end
            end
            if ~isempty(rulesetTotal{i,j,k}) 
                if rulesetTotal{i,j,k}~=-2
                    if j == AntMinerPos || j == RipperPos % ignore defaultrule for attribute tests
                        nrTestsRuleTotalArray = countTestsRuleset(rulesetTotal{i,j,k}(1:end-1,:),rulesetTotal{i,j,k}(end,:),variableType{i,1});
                        nrTestsRuleTotal(i,j,k) = mean(nrTestsRuleTotalArray,1); 
                        nrRulesTotal(i,j,k) = size(rulesetTotal{i,j,k},1);
                    else % if j == J48Pos
                        nrTestsRuleTotalArray = countTestsRuleset(rulesetTotal{i,j,k}(1:end,:),rulesetTotal{i,j,k}(end,:),variableType{i,1},adjValue);
                        nrTestsRuleTotal(i,j,k) = mean(nrTestsRuleTotalArray,1); 
                        nrRulesTotal(i,j,k) = size(rulesetTotal{i,j,k},1); 
                    end
                end
            end
        end
    end
end

%% Save the results of the experiments
save 'RulemBenchmarkingExperimentResults.mat'


