%% Author: Wouter Verbeke
% Data Analytics Laboratory
% Department of Business Technology and Operations 
% Faculty of Economic and Social Sciences and Solvay Business School
% Vrije Universiteit Brussel, Pleinlaan 2, B-1050 Brussel, Belgium

%% Documentation
% Evaluate rule set on 'a' data set (test data for instance)
% The rule set is in the format defined by the AntMiner+ implementation of
% Bart Minnaert, this code is based as well on a part of the AntMinerCore
% matlab file of the AntMiner+ implementation 

function [accuracy,positive,notCovered,list,mae,mse] = evalRulesetMaeMse(ruleset,data,valueTypes,anyValue,classIdx,rulesetClassIdx,suppressOutput)

% ruleset contains the rules
% data contains the instances, with the rows the observations and the
% columns the attributes
% valueTypes indicates the type of variable in a column: nominal (0) or
% ordinal (1)
% anyValue is the value in the ruleset representing "any"
% classIdx is an optional argument that indicates the column with the
% target variable
% values is a mapping function between the data labels and the values used
% in the rules

if nargin < 5 % then class index is not passed, and we assume that the last column in the data contains the class
    classIdx = size(data,2);
    rulesetClassIdx = size(ruleset,2);
end
if nargin < 7
    suppressOutput = 0;
end

%% Evaluate rule set accuracy on test set
nextRuleIndex = size(ruleset,1); % the number of rules in the ruleset
nbVariables = size(valueTypes,2); % the number of variables in the data set
nbTest = size(data,1);
positive = 0;
notCovered = 0;
list = [];
squarederror = 0;
abserror = 0;

for i = 1:nbTest %cover all datapoints in the test set
    covered = 0; %indicates it is covered by a rule in the list
    for j = 1:nextRuleIndex %try each rule in order
        condition = 1;
        idx = 1;
        for k = 1:nbVariables %check if the datapoint matches the rule
            if valueTypes(k) == 0 %nominal variable
                if ruleset(j,idx) ~= data(i,k) && ruleset(j,idx) ~= anyValue %take into account the 'any'-case
                    condition = 0;
                    break;
                end
                idx = idx + 1;
            end
            if valueTypes(k) == 1 %ordinal variable => lower and upper bound
                if ruleset(j,idx) > data(i,k) %lower bound restriction
                    condition = 0;
                    break;
                end
                if ruleset(j,idx+1) <= data(i,k) %upper bound restriction
%                 if ruleset(j,idx+1) < data(i,k) %upper bound restriction
                    condition = 0;
                    break
                end
                idx = idx + 2; %skip the 2 nominal variables of the ordinal variable
            end
        end
        if condition == 1 %datapoint matches the j-th rule in the decision list
            if ruleset(j,rulesetClassIdx) == data(i,classIdx) %good prediction
                positive = positive + 1;
                list = [list;i];
            else
                diff = ruleset(j,rulesetClassIdx)-data(i,classIdx);
                abserror = abserror + abs(diff);
                squarederror = squarederror+diff*diff;
            end
            covered = 1;
            break;
        end
    end
    if covered == 0
        notCovered = notCovered + 1;
        if ~suppressOutput
            disp('!!! Warning: instance not covered by ruleset! !!!');
        end
    end
end
accuracy = positive/nbTest;
mae = abserror/nbTest;
mse = squarederror/nbTest;
