% function [allData_trn_clust,allData_tst_clust,clust] = fn_cluster(...
% .. allData_trn,allData_tst,nrVariable_array,nrClusters_array,...
%  csv_outputnames,newFilename_trn,newFilename_tst,stringVals,featureNames)
% Thomas Verbraken - 03/30/2010 - Katholieke Universiteit Leuven
%
% This function clusters the values a categorical variable (nrVariable in
% the allData matrix) can take into nrClusters clusters. It is able to
% handle a dataset split into a training and test set, so that it applies
% the clustering algorithm only on the training set, and afterwards applies
% the outcome to both training and test set. The output clust contains the
% initial variable values and to which cluster (0..nrClusters-1) they are 
% assigned.

% The function can cluster more than one variable, i.e. nrVariable_array
% can contain more than one variable. Afterwards the function writes the
% output to an arff file, if the last four inputs are specified.
%
% The data matrix is a matrix odds with odds(i,j) = P(class=j|var=i) with
% j=1..Nclasses-1. Certain values of the variable which are close to each
% other in terms of above conditional probabilities, have high chances of
% predicting the same class. The euclidiean distance measure is applied.

function [allData_trn,allData_tst,clust] = fn_cluster(allData_trn,allData_tst,nrVariable_array,nrClusters_array,csv_outputnames,varargin)
if ~isempty(varargin)
    newFilename_trn = varargin{1};
    newFilename_tst = varargin{2};
    stringVals = varargin{3};
    featureNames = varargin{4};
end

for n=1:length(nrVariable_array)
    nrVariable = nrVariable_array(n);
    nrClusters = nrClusters_array(n);
    %% load data
    mdata = [allData_trn(:,nrVariable) allData_trn(:,end)];
    data_trn_clust = nan(size(allData_trn,1),1);
    data_tst_clust = nan(size(allData_tst,1),1);
    %% Calculate the odds = P(c=j|var=i) for each variable i and class j
    var_val = unique(mdata(:,1));
    class_val = unique(mdata(:,2));
    countsVar = zeros(length(var_val),1);
    countsClass = zeros(length(var_val),length(class_val));
    for i=1:length(var_val)
        countsVar(i) = sum(mdata(:,1)== var_val(i));
        for j=1:length(class_val)
            countsClass(i,j) = sum((mdata(:,1)==var_val(i)) .* (mdata(:,2)==class_val(j)));
        end
    end
    odds = (countsClass(:,1:end-1) ./ (countsVar * ones(1,length(class_val)-1)));

    %% check whether some odds values are incorrect (0 or 1)
    % 0 values replaced with minimum, 1 values with maximum
    for i=1:size(odds,2)
        odds(odds(:,i)==0,i) = min(odds(odds(:,i)~=0));
        odds(odds(:,i)==1,i) = max(odds(odds(:,i)~=1));
    end

    %% Perform hierarchical agglomerative clustering to end up with four
    %% clusters of values of the categorical variable
    %% We use function agglom.m and related functions
    Y = pdist(odds,'euclid'); 
    Z = linkage(Y,'single'); 
    T = cluster(Z,'maxclust',nrClusters); 
    clust = [var_val T-1];
    % Output
    for i=1:length(clust(:,1))
        data_trn_clust(allData_trn(:,nrVariable)==clust(i,1)) = clust(clust(:,1)==clust(i,1),2);
    end
    if ~isempty(allData_tst)
        var_val_ext = union(var_val,unique(allData_tst(:,nrVariable)));
        clust_ext = [var_val_ext nan(size(var_val_ext))];
        for j=1:length(clust(:,1))
            clust_ext(clust_ext(:,1)==clust(j,1),2) = clust(j,2);
        end
        clust_ext(isnan(clust_ext(:,2)),2) = mode(clust(:,2));
       for i=1:length(clust_ext(:,1))
            data_tst_clust(allData_tst(:,nrVariable)==clust_ext(i,1)) = clust_ext(clust_ext(:,1)==clust_ext(i,1),2);
       end
       clust = clust_ext;
    end
    if sum( [isnan(data_trn_clust) ; isnan(data_tst_clust)]) > 0
        error('Not all values are assigned to a cluster')
    end
    allData_trn(:,nrVariable) = data_trn_clust;
    allData_tst(:,nrVariable) = data_tst_clust;
    clust_name = cell(length(stringVals{nrVariable}),3);
    % Write cluster information to csv file
    if ~isempty(csv_outputnames)
        fid = fopen(csv_outputnames{n},'w');
        for i=1:length(stringVals{nrVariable})
            clust_name(i,:) = {i-1 stringVals{nrVariable}(i) clust(clust(:,1)==i-1,2)};
            fprintf(fid,'%s',[int2str(i-1) ',' char(stringVals{nrVariable}(i)) ',' int2str(clust(clust(:,1)==i-1,2)) ]);
            fprintf(fid,'%s\n','');
        end
        fclose(fid);
    end
end
if ~isempty(varargin)
    % Write to arff file
    nom_type = [];
    for i=1:length(stringVals)
        if ~isempty(stringVals{i})
            nom_type = [nom_type i];
        end
    end
    newWekaOBJ = matlab2weka('SplitData_missValues_clustered', featureNames, allData_trn);
    idx_arff = strfind(newFilename_trn,'.arff');
    saveARFF([newFilename_trn(1:idx_arff-1) '_TEMP.arff'],newWekaOBJ);
    if ~isempty(allData_tst)
        newWekaOBJ = matlab2weka('SplitData_missValues_clustered', featureNames, allData_tst);
        idx_arff = strfind(newFilename_tst,'.arff');
        saveARFF([newFilename_tst(1:idx_arff-1) '_TEMP.arff'],newWekaOBJ);
        fn_convert_num2nom([newFilename_trn(1:idx_arff-1) '_TEMP.arff'],newFilename_trn,nom_type,[newFilename_tst(1:idx_arff-1) '_TEMP.arff'],newFilename_tst);
    else
        fn_convert_num2nom([newFilename_trn(1:idx_arff-1) '_TEMP.arff'],newFilename_trn,nom_typ)
    end
end

if isempty(nrVariable_array)
    clust = [];
end
