% function [stringVals] = fn_discrFayyad(orFilename_trn, ...
%      ... newFilename_trn,orFilename_tst,newFilename_tst, attr)
% Thomas Verbraken - 03/30/2010 - Katholieke Universiteit Leuven
%
% This function discretizes continuous attributes in an original arrf file.
% It is possible to only apply this on a training set, or on a training and
% test set (test set follows same discretization rules as the training
% set). The output is an arff file, but the information about the intervals
% is not stored in the arff file itself, however it is obtainable as matlab
% output in the form of the stringVals array. This function also removes
% useless attributes (i.e. flat attributes).
%
% For flat attributes which are deleted, a Fisher and Pearson test is
% conducted prior to discretization. This information is available in a 
% spreadsheet.


function [stringVals useless] = fn_discrFayyad(orFilename_trn,newFilename_trn,orFilename_tst,newFilename_tst,varargin)

%% load data
wekaOBJ_trn = loadARFF(orFilename_trn);
[allData_trn_orig,featureNames_orig,~,stringVals,~] = weka2matlab(wekaOBJ_trn,[]);
if ~isempty(varargin)
    attr = varargin{1};
else
    attr = [];
    for i=1:length(stringVals)
        if isempty(stringVals{i})
            attr = [attr i];
        end
    end
end
opt = '';
for i=1:length(attr)
    if isempty(opt)
        opt = int2str(attr(i));
    else
        opt = [opt ',' int2str(attr(i))];
    end
end
% Apply filter
filter = javaObject('weka.filters.supervised.attribute.Discretize');
filter.setOptions(opt);
filter.setInputFormat(wekaOBJ_trn);
wekaOBJ_trn_discr = filter.useFilter(wekaOBJ_trn, filter);
[allData_trn,featureNames,~,stringVals,~] = weka2matlab(wekaOBJ_trn_discr,[]);
% Check which attributes are useles, i.e. flat
useless = [];
for i=1:length(stringVals)
    if length(stringVals{i})==1
        useless = [useless i];
    end
end
% Save training file in arff format - useless attributes are removed
allData_trn_output = allData_trn(:,setdiff(1:size(allData_trn,2),useless));
featureNames = featureNames(setdiff(1:size(allData_trn,2),useless));
wekaOBJ_trn_discr = matlab2weka('SplitData_missValues_clustered_discret', featureNames, allData_trn_output);
idx_arff = strfind(newFilename_trn,'.arff');
saveARFF([newFilename_trn(1:idx_arff-1) '_TEMP.arff'],wekaOBJ_trn_discr);
clear wekaOBJ_trn wekaObj_trn_discr

% check if there is a test file and apply the same discretization principle
if ~isempty(orFilename_tst)
    wekaOBJ_tst = loadARFF(orFilename_tst);
    [allData_tst,~,~,~,~] = weka2matlab(wekaOBJ_tst,[]);
    allData_tst_output = allData_tst;
    allData_tst_output(:,attr) = NaN;
    for i=1:length(attr)
        iv = [];
        if ~isempty(stringVals{attr(i)})
            for j=1:length(stringVals{attr(i)})
                val = char(stringVals{attr(i)}(j));
                all_true = ~isempty(strfind(val,'All'));
                init = strfind(val,'(');
                midl = strfind(val,'-');
                if length(midl) > 1
                    midl = midl(2);
                end
                if all_true
                    iv = [-inf];
                else
                    iv = [iv str2double(val(init+1:midl-1))];
                end
            end
        end
        allData_tst_output(:,attr(i)) = sum(ones(size(allData_tst,1),1)*iv < allData_tst(:,attr(i))*ones(1,length(iv)),2)-1;        
    end
    allData_tst_output = allData_tst_output(:,setdiff(1:size(allData_tst,2),useless));
    wekaOBJ_tst_discr = matlab2weka('SplitData_missValues_clustered_discret', featureNames, allData_tst_output);
    idx_arff = strfind(newFilename_tst,'.arff');
    saveARFF([newFilename_tst(1:idx_arff-1) '_TEMP.arff'],wekaOBJ_tst_discr);
    fn_convert_num2nom([newFilename_tst(1:idx_arff-1) '_TEMP.arff'],newFilename_tst,1:size(allData_tst,2))
end
stringVals = stringVals(setdiff(1:size(allData_trn,2),useless));
if isempty(orFilename_tst)
    fn_convert_num2nom([newFilename_trn(1:idx_arff-1) '_TEMP.arff'],newFilename_trn,1:size(allData_trn,2))
else
    fn_convert_num2nom([newFilename_trn(1:idx_arff-1) '_TEMP.arff'],newFilename_trn,1:size(allData_trn,2),[newFilename_tst(1:idx_arff-1) '_TEMP.arff'],newFilename_tst)
end
    
%% If there are items removed, the function writes a file with the Fisher
% scores for the removed items
if ~isempty(useless)
    warning(['A number of variables was flat after discretization (i.e. 1 bin) - Fisher test and Pearson statistic prior to discretization can be found in file ' newFilename_trn(1:idx_arff-1) '_FisherPearsonRemoved.xls' ]);
    x = allData_trn_orig(:,useless);
    N = size(x,1);
    featureNames_x = featureNames_orig(useless);
    y = allData_trn(:,end);
    x_c  = x(y==1,:);
    x_nc = x(y==0,:);
    Fs = abs(mean(x_c,1)'-mean(x_nc,1)')./sqrt(std(x_c,1)'.^2 + std(x_nc,1)'.^2);
    r_p = ((x-ones(size(x,1),1)*mean(x,1))' * (y-mean(y,1))) ./ ((N-1)* std(x,1)' * std(y,1));
    r_p = abs(r_p);
    output = [num2cell(useless') featureNames_x' num2cell(Fs) num2cell(r_p)];
    header = {'Attr #' 'Attr Name' 'Fisher Score' 'Pearson Statistic'};
    output = [header;output];
    % write output to file
    xlswrite([newFilename_trn(1:idx_arff-1) '_FisherPearsonRemoved.xls' ],output);
end
    
