% function [allData_trn,allData_tst]=fn_replaceMissing_fromMAT( ...
%  ... allData_trn,allData_tst,stringVals,mode_imp,median_imp,mean_imp)
% Thomas Verbraken - 03/29/2010 - Katholieke Universiteit Leuven
%
% This function replaces missing values in a given training and test set.
% Optional input arguments are mode_imp, median_imp and mean_imp. Whether
% all three should be defined or none, and they contain the attributes
% to which the respective imputation should be applied. If left blank, mode
% imputation is applied for all nominal attributes, and median imputation
% for all numeric attributes.
% Outputs are the training and test set with replaced values and a matrix
% indicating the percentage of missing values in traning and test set for 
% each attribute.

function [allData_trn,allData_tst,missVal]=fn_replaceMissing_fromMAT(allData_trn,allData_tst,stringVals,varargin)

if isempty(varargin)
    warning('No specification given: median imputation for all numeric attributes; mode imputation for all nominal attributes')
else if length(varargin) == 3
        mode_imp = varargin{1};
        median_imp = varargin{2};
        mean_imp = varargin{3};
    else
        error('Incorrect numer of arguments')
    end
end

%% check which attributes are nominal/numeric
nom_type = [];
for i=1:length(stringVals)
    if ~isempty(stringVals{i})
        nom_type = [nom_type i];
    end
end
num_type = setdiff(1:size(stringVals,2),nom_type);
% set the default imputation values
if isempty(varargin)
    median_imp = num_type;
    mode_imp = nom_type;
    mean_imp = [];
end 
%% Percentage missing values for each attribute
missVal = zeros(size(allData_trn,2),3);
for i=1:size(allData_trn,2)
    missVal(i,1) = i;
    missVal(i,2) = sum(isnan(allData_trn(:,i)))/size(allData_trn,1)*100;
    missVal(i,3) = sum(isnan(allData_tst(:,i)))/size(allData_tst,1)*100;
end

%% imputation
% mode imputation
for j = 1:length(mode_imp)
    attr_i = mode_imp(j);
    idx_miss = find(isnan(allData_trn(:,attr_i)));
    idx_notmiss = setdiff(1:size(allData_trn,1),idx_miss);
    allData_trn(idx_miss,attr_i) = mode(allData_trn(idx_notmiss,attr_i),1);
    if ~isempty(allData_tst)
        allData_tst(isnan(allData_tst(:,attr_i)),attr_i) = mode(allData_trn(idx_notmiss,attr_i),1);
    end
end
if isempty(mode_imp)
    warning('No mode imputation applied')
end
        
% median imputation
for j = 1:length(median_imp)
    attr_i = median_imp(j);
    idx_miss = find(isnan(allData_trn(:,attr_i)));
    idx_notmiss = setdiff(1:size(allData_trn,1),idx_miss);
    allData_trn(idx_miss,attr_i) = median(allData_trn(idx_notmiss,attr_i),1);
    if ~isempty(allData_tst)
        allData_tst(isnan(allData_tst(:,attr_i)),attr_i) = median(allData_trn(idx_notmiss,attr_i),1);
    end
end
if isempty(median_imp)
    warning('No median imputation applied')
end

% mean imputation
for j = 1:length(mean_imp)
    attr_i = mean_imp(j);
    idx_miss = find(isnan(allData_trn(:,attr_i)));
    idx_notmiss = setdiff(1:size(allData_trn,1),idx_miss);
    allData_trn(idx_miss,attr_i) = mean(allData_trn(idx_notmiss,attr_i),1);
    if ~isempty(allData_tst)
        allData_tst(isnan(allData_tst(:,attr_i)),attr_i) = mean(allData_trn(idx_notmiss,attr_i),1);
    end
end
if isempty(mean_imp)
    warning('No mean imputation applied')
end
