% function fn_splitDataset(orFilename,perc_train,stratif_true,newFilename)
% Thomas Verbraken - 03/29/2010 - Katholieke Universiteit Leuven
%
% The function reads an original arff file and splits it into a training
% set and a test set, with perc_train (expressed as fraction) being in the
% training set. With boolean stratif_true equal to 1, stratified sampling
% will be done (with regards to the target variable, the last variable in
% the arff file). The input newFilename is optional. When a file name is
% given, the function will also write the training and test set to an arff
% file. However, this has one disadvantage: a nominal string variable will
% be converted to a number in the weka output file. The information in the 
% string itself is lost. 

function [allData_trn,allData_tst,featureNames,stringVals,nom_type]=fn_splitDataset(orFilename,perc_train,stratif_true,varargin)
  
%% Connect to weka and load file
wekaOBJ = loadARFF(orFilename);
[allData,featureNames,~,stringVals,~] = weka2matlab(wekaOBJ,[]);
N_instanc = size(allData,1);
nom_type = [];
for i=1:length(stringVals)
    if ~isempty(stringVals{i})
        nom_type = [nom_type i];
    end
end

%% Select instances for the training and test set
if stratif_true == 1
    idx_train = fn_stratSample(allData(:,end),round(perc_train*N_instanc));
    idx_test = setdiff(1:N_instanc,idx_train);
else
    idx_train = randsample(N_instanc,round(perc_train*N_instanc));
    idx_test = setdiff(1:N_instanc,idx_train);
end
allData_trn = allData(idx_train,:);
allData_tst = allData(idx_test,:);

%% testing if all possible values for the discrete variables are present in
% the training set.
for i=1:length(nom_type)
    varVal = setdiff(unique(allData(:,nom_type(i))),NaN);
    varVal = varVal(~isnan(varVal));
    for j=1:length(varVal)
        if sum(allData_trn(:,nom_type(i))==varVal(j))==0
            nr_ind = find(allData_tst(:,nom_type(i))==varVal(j));
            idx1 = nr_ind(ceil(rand*length(nr_ind)));
            nr_ind = find(allData_trn(:,end)==allData_tst(idx1,end));
            if isempty(nr_ind)
                nr_ind = 1:size(allData_trn,1);
            end
            idx2 = nr_ind(ceil(rand*length(nr_ind)));
            idx_trn = setdiff(1:size(allData_trn,1),idx2);
            allData_trn_temp = allData_trn;
            allData_trn_temp(idx2,isnan(allData_trn(idx2,:))) = 0;
            swap_ok = sum(sum(allData_trn(idx_trn,nom_type)==(ones(length(idx_trn),1)*allData_trn_temp(idx2,nom_type)),1)==0)==0;
            while and(~swap_ok,~isempty(nr_ind))
                nr_ind = setdiff(nr_ind,idx2);
                idx2 = nr_ind(ceil(rand*length(nr_ind)));
                idx_trn = setdiff(1:size(allData_trn,1),idx2);
                allData_trn_temp = allData_trn;
                allData_trn_temp(idx2,isnan(allData_trn(idx2,:))) = 0;
                swap_ok = sum(sum(allData_trn(idx_trn,nom_type)==(ones(length(idx_trn),1)*allData_trn_temp(idx2,nom_type)),1)==0)==0;
            end
            if ~swap_ok
                error('Problem with the split into training and test set. Discrete variables do not allow that all possible values are represented in the training set.');
            end
            idx_tst = setdiff(1:size(allData_tst,1),idx1);
            allData_trn = [allData_trn(idx_trn,:);allData_tst(idx1,:)];
            allData_tst = [allData_tst(idx_tst,:);allData_trn(idx1,:)];
        end
    end
end
%% Test if all possible values for all variables are present in the
% training set. 
for i=1:length(nom_type)
    varVal = unique([allData_trn(:,nom_type(i));allData_tst(:,nom_type(i))]);
    varVal = varVal(~isnan(varVal));
    for j=1:length(varVal)
        if sum(allData_trn(:,nom_type(i))==varVal(j))==0
            warning('Problem with split of training and test set! Not all possible values for variables are present in training set.')
        end
    end
end


%% Store output
if ~isempty(varargin)
    newFilename = varargin{1};
    idx_arff = strfind(newFilename,'.arff');
    if isempty(idx_arff)
        error('The newFilename is not an arff file!')
    else
        newFilename_trn = [newFilename(1:idx_arff-1) '_trn.arff'];
        newFilename_tst = [newFilename(1:idx_arff-1) '_tst.arff'];
    end
    newWekaOBJ = matlab2weka('splitData', featureNames, allData_trn);
    saveARFF([newFilename(1:idx_arff-1) '_trn_TEMP.arff'],newWekaOBJ);
    newWekaOBJ = matlab2weka('splitData', featureNames, allData_tst);
    saveARFF([newFilename(1:idx_arff-1) '_tst_TEMP.arff'],newWekaOBJ);
    fn_convert_num2nom([newFilename(1:idx_arff-1) '_trn_TEMP.arff'],newFilename_trn,nom_type,[newFilename(1:idx_arff-1) '_tst_TEMP.arff'],newFilename_tst);
end
