%% Author: Wouter Verbeke
% Data Analytics Laboratory
% Department of Business Technology and Operations 
% Faculty of Economic and Social Sciences and Solvay Business School
% Vrije Universiteit Brussel, Pleinlaan 2, B-1050 Brussel, Belgium

%% Documentation
% 'data file' is an arff file without missing values, with only discrete
% ordinal or nominal variables,  and all typical preprocessing steps 
% executed. Attributes are indicated by their first value to be ordinal or
% nominal. The output of this function:
% wekadata:     the data set
% valueTypes:   vector with value types of the variables:
%                   0 nominal
%                   1 ordinal
% maxValues:    the largest number of values found for a variable => will
%               be used for the dimensions of all vectors/matrices that
%               work with this.
% values:       vector containing a vector of values for each variable
% nbValues:     contains the number of values for each variable

function [wekadata,valueTypes,maxValues,values,nbValues] = readData(datafile)

%% read data using weka
import weka.core.Instances;
import java.io.InputStreamReader;
import java.io.FileInputStream;
reader = InputStreamReader(FileInputStream(datafile));
wekadata = Instances(reader);
reader.close();

%% classIndex: find the class attribute and switch it so it is the last
%% attribute in the data
classFound = 0;
classIndex = -1;
for i=0:wekadata.numAttributes-1
    if wekadata.attribute(i).isNominal == 1
        if strcmp(wekadata.attribute(i).value(0),'class') == 1
            if classFound == 1
                fprintf('ERROR: multiple class attributes found in input');
                return;
            else
                classFound = 1;
            end
            classIndex = i;
        end
    end
end
if classIndex == -1
    fprintf('ERROR: no class attribute found in input');
    return;
end
import weka.filters.unsupervised.attribute.Reorder
reorderFilter = Reorder();
clear order;
for iWeka = 0:wekadata.numAttributes-1
    i = iWeka+1;
    if iWeka < classIndex
        order(i) = iWeka;
    elseif iWeka < wekadata.numAttributes-1
        order(i) = iWeka+1;
    else
        order(i) = classIndex;
    end
end
reorderFilter.setAttributeIndicesArray(order);
reorderFilter.setInputFormat(wekadata);
wekadata = weka.filters.Filter.useFilter(wekadata,reorderFilter);

%% valueTypes, nbValues, maxValues, values
maxValues = 1; %starts at 3 because of alpha and beta
for iWeka = 0:wekadata.numAttributes-2 %last variable is class variable
    i = iWeka + 1; %iWeka is the correct index for the weka-objects, i is the correct index for the matlab structs
    if wekadata.attribute(iWeka).isNominal == 1
        if strcmp(wekadata.attribute(iWeka).value(0),'ordinal')
            valueTypes(i) = 1;
        elseif strcmp(wekadata.attribute(iWeka).value(0),'nominal')
            valueTypes(i) = 0;
        else
            disp('Error: No value type assigned in data file')
            return
        end
    else
        disp('Error: numeric attribute in data file');
        return
    end
    nbValues(i) = wekadata.attribute(iWeka).numValues();
    if nbValues(i) > 0 % account for the first value that indicates it is nominal or ordinal
        nbValues(i) = nbValues(i)-1;
    else
        values(i,1) = -1; %BUGFIX: making sure the row is initialised. a value '-1' after preprocessing should not occur.
    end
    if nbValues(i) > maxValues
        maxValues = nbValues(i);
    end
    % the transformations of the weka from symbolic to numeric will
    % give each value a number starting from 1
    for j= 1:nbValues(i)                
        values(i,j) = j;               
    end
end

%% remove the first attribute value for all nominal and ordinal
%% attributes (this value is either 'NOMINAL' or 'ORDINAL')
%% value removing in weka: merge the unused value with another value
%% rename again because it also merges the names
import weka.filters.unsupervised.attribute.MergeTwoValues
wekadata.setClassIndex(-1); %weka doesn't like merging class values
for i=0:wekadata.numAttributes()-1
   prevvalue = wekadata.attribute(i).value(1);              
   valueRemover = MergeTwoValues();
   valueRemover.setAttributeIndex(num2str(i+1));
   valueRemover.setFirstValueIndex('1');
   valueRemover.setSecondValueIndex('2');
   valueRemover.setInputFormat(wekadata);
   wekadata = weka.filters.Filter.useFilter(wekadata,valueRemover);   
   wekadata.renameAttributeValue(i,0,prevvalue);
end
wekadata.setClassIndex(wekadata.numAttributes()-1);

%% convert java object to data using weka2matlab
wekadata = weka2matlab(wekadata);

